//
// includes
// --------
#define _DEFAULT_SOURCE
#define _BSD_SOURCE
#define _GNU_SOURCE

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include "config.h"
//
// definitions
// -----------
#define CSEDIT_VERSION "0.0.1"
#define CTRL_KEY(k) ((k) & 0x1f)
#define HL_HIGHLIGHT_NUMBERS (1<<0)

// 
// definitions for config file
// ---------------------------

enum editorMode {CommandMode, InputMode, VisualMode};

enum editorHighlight {
  HL_NORMAL,
  HL_NUMBER
};

enum editorKey {
  BACKSPACE = 127,
  ARROW_LEFT = 1000,
  ARROW_RIGHT,
  ARROW_UP,
  ARROW_DOWN,
  PAGE_UP,
  PAGE_DOWN,
  DEL_KEY,
  HOME_KEY,
  END_KEY
};

//
// data
// ----

typedef struct erow {
  int size;
  int rendersize;
  char *chars;
  char *render;
  unsigned char *hl;
} erow;

struct editorConfig {
  enum editorMode mode;
  int rowoff;
  int coloff;
  int cx, cy, rx;
  int screenRows;
  int screenCols;
  int numRows;
  int unsaved;
  erow *row;
  char *filename;
  char statusmsg[80];
  time_t statusmsg_time;
  struct termios orig_termios;
  struct editorSyntax *syntax;
};

struct editorSyntax {
  char *filetype;
  char **filematch;
  int flags;
};

struct editorConfig E;


// 
// filetypes
// ---------
char *C_HL_extensions[] = {".c", ".h", ".cpp", NULL};

struct editorSyntax HLDB[] = {
  {
    "c",
    C_HL_extensions,
    HL_HIGHLIGHT_NUMBERS
  },
};

#define HLDB_ENTRIES (sizeof(HLDB) / sizeof(HLDB[0]))

// 
// prototypes
// ----------
void editorStatusMessage(const char *fmt, ...);
void editorRefreshScreen();
char *editorPrompt(char *prompt);

//
// terminal
// --------
void die(const char *s) {
  write(STDOUT_FILENO, "\033[2J", 4);
  write(STDOUT_FILENO, "\033[H", 3);
  perror(s);
  write(STDOUT_FILENO, "\r", 1);
  exit(1);
}

void quit() {
  write(STDOUT_FILENO, "\033[2J", 4);
  write(STDOUT_FILENO, "\033[H", 3);
  exit(0);
}

void disableRawMode() {
  if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &E.orig_termios) == -1) 
    die("tcsetattr");
}

void enableRawMode() {
  if (tcgetattr(STDIN_FILENO, &E.orig_termios) == -1) die("tcgetattr");
  atexit(disableRawMode);

  struct termios raw = E.orig_termios;
  raw.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  raw.c_oflag &= ~(OPOST);
  raw.c_cflag |= ~(CS8);
  raw.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
  raw.c_cc[VMIN] = 0;
  raw.c_cc[VTIME] = 1;

  if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw) == -1) die("tcsetattr");
}

int editorReadKey() {
  int nread;
  char c;
  while ((nread = read(STDIN_FILENO, &c, 1)) != 1) {
    if (nread == -1 && errno != EAGAIN) die("read");
  }

  if (c == '\033') {
    char seq[3];

    if (read(STDIN_FILENO, &seq[0], 1) != 1) return '\033';
    if (read(STDIN_FILENO, &seq[1], 1) != 1) return '\033';
    if (seq[0] == '[') {
      if (seq[1] >= '0' && seq[1] <= '9') {
        if (read(STDIN_FILENO, &seq[2], 1) != 1) return '\033';
        if (seq[2] == '~') {
          switch (seq[1]) {
            case '1': return HOME_KEY;
            case '3': return DEL_KEY;
            case '4': return END_KEY;
            case '5': return PAGE_UP;
            case '6': return PAGE_DOWN;
            case '7': return HOME_KEY;
            case '8': return END_KEY;
          }
        }
      } else {
        switch (seq[1]) {
          case 'A': return ARROW_UP;
          case 'B': return ARROW_DOWN;
          case 'C': return ARROW_RIGHT;
          case 'D': return ARROW_LEFT;
          case 'H': return HOME_KEY;
          case 'F': return END_KEY;
        }
      }
    } else if (seq[0] == 'O') {
      switch (seq[1]) {
        case 'H': return HOME_KEY;
        case 'F': return END_KEY;
      }
    }
    return '\033';
  }
  return c;
}

int getCursorPosition(int *rows, int *cols) {
  char buf[32];
  unsigned int i = 0;

  if (write(STDOUT_FILENO, "\033[6n", 4) != 4) return -1;
  printf("\r\n");

  while (i < sizeof(buf) - 1) {
    if (read(STDIN_FILENO, &buf[i], 1) != 1) break;
    if (buf[i] == 'R') break;
    i++;
  }
  buf[i] = '\0';
  if (buf[0] != '\033' || buf[1] != '[') return -1;
  if (scanf(&buf[2], "%d;%d", rows, cols) != 2) return -1;
  return 0;
}

int getWindowSize(int *rows, int *cols) {
  struct winsize ws;

  if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) == -1 || ws.ws_col == 0) {
    if (write(STDOUT_FILENO, "\033[999C\033[999B", 12) != 12) return -1;
    return getCursorPosition(rows, cols);
  }

  *cols = ws.ws_col;
  *rows = ws.ws_row;
  return 0;
}

// 
// syntax highlighting
// -------------------
int is_separator(int c) {
  return isspace(c) || c == '\0' || strchr(",.()+-/*=~%<>[];", c) != NULL;
}

void editorUpdateSyntax(erow *row) {
  row->hl = realloc(row->hl, row->rendersize);
  memset(row->hl, HL_NORMAL, row->rendersize);

  if (E.syntax == NULL) return;

  int prev_sep = 1;

  int i;
  while (i < row->size) {
    char c = row->render[i];
    unsigned char prev_hl = (i > 0) ? row->hl[i - 1] : HL_NORMAL;

    if (E.syntax->flags & HL_HIGHLIGHT_NUMBERS) {
      if ((isdigit(c) && (prev_sep || prev_hl == HL_NUMBER)) || 
          (c == '.' && prev_hl == HL_NUMBER)) {
        row->hl[i] = HL_NUMBER;
        i++;
        prev_sep = 0;
        continue;
      }
    }
    prev_sep = is_separator(c);
    i++;
  }
}

int editorSyntaxToColor(int hl) {
  switch (hl) {
    case HL_NUMBER: return HL_NUMBER_COLOR;
    default: return HL_DEFAULT_COLOR;
  }
}

void editorSelectSyntaxHighlight() {
  E.syntax = NULL;
  if (E.filename == NULL) return;

  char *ext = strrchr(E.filename, '.');

  for (unsigned int j = 0; j < HLDB_ENTRIES; j++) {
    struct editorSyntax *s = &HLDB[j];
    unsigned int i = 0;
    while (s->filematch[i]) {
      int is_ext = (s->filematch[i][0] == '.');
      if ((is_ext && ext && !strcmp(ext, s->filematch[i])) ||
          (!is_ext && strstr(E.filename, s->filematch[i]))) {
        E.syntax = s;
        return;
      }
      i++;
    }
  }
}

//
// row ops
// -------
int editorRowCxToRx(erow *row, int cx) {
  int rx = 0;
  int j;
  for (j = 0; j < cx; j++) {
    if (row->chars[j] == '\t')
      rx += (tabwidth - 1) - (rx % tabwidth);
    rx++;
  }
  return rx;
}

void editorUpdateRow(erow *row) {
  int tabs = 0;
  int j;
  for (j = 0;  j < row->size; j++)
    if (row->chars[j] == '\t') tabs++;

  free(row->render);
  row->render = malloc(row->size + tabs * (tabwidth - 1) + 1);

  int idx = 0;
  for (j = 0; j < row->size; j++) {
    if (row->chars[j] == '\t') {
      row->render[idx++] = ' ';
      while (idx % tabwidth != 0) row->render[idx++] = ' ';
    } else {
      row->render[idx++] = row->chars[j];
    }
  }
  row->render[idx] = '\0';
  row->rendersize = idx;

  editorUpdateSyntax(row);
}

void editorInsertRow(int at, char *s, size_t len) {
  if (at < 0 || at > E.numRows) return;

  E.row = realloc(E.row, sizeof(erow) * (E.numRows + 1));
  memmove(&E.row[at + 1], &E.row[at], sizeof(erow) * (E.numRows - at));

  E.row[at].size = len;
  E.row[at].chars = malloc(len + 1);
  memcpy(E.row[at].chars, s, len);
  E.row[at].chars[len] = '\0';

  E.row[at].rendersize = 0;
  E.row[at].render = NULL;
  E.row[at].hl = NULL;
  editorUpdateRow(&E.row[at]);

  E.numRows++;
  E.unsaved++;
}

void editorFreeRow(erow *row) {
  free(row->render);
  free(row->chars);
  free(row->hl);
}

void editorDelRow(int at) {
  if (at < 0 || at >= E.numRows) return;
  editorFreeRow(&E.row[at]);
  memmove(&E.row[at], &E.row[at + 1], sizeof(erow) * (E.numRows - at - 1));
  E.numRows--;
  E.unsaved++;
}

void editorRowInsertChar(erow *row, int at, int c) {
  if (at < 0 || at > row->size) at = row->size;
  row->chars = realloc(row->chars, row->size+2);
  memmove(&row->chars[at + 1], &row->chars[at], row->size - at + 1);
  row->size++;
  row->chars[at] = c;
  editorUpdateRow(row);
  E.unsaved++;
}

void editorRowAppendString(erow *row, char *s, size_t len) {
  row->chars = realloc(row->chars, row->size + len + 1);
  memcpy(&row->chars[row->size], s, len);
  row->size += len;
  row->chars[row->size] = '\0';
  editorUpdateRow(row);
  E.unsaved++;
}

void editorRowDeleteChar(erow *row, int at) {
  if (at < 0 || at >= row->size) return;
  memmove(&row->chars[at], &row->chars[at + 1], row->size - at);
  row->size--;
  editorUpdateRow(row);
  E.unsaved++;
}

// 
// editor ops
// ----------
void editorInsertChar(int c) {
  if (E.cy ==  E.numRows) {
    editorInsertRow(E.numRows, "", 0);
  }
  editorRowInsertChar(&E.row[E.cy], E.cx, c);
  E.cx++;
}

void editorDeleteChar() {
  if (E.cy == E.numRows) return;
  if (E.cy == 0 && E.cx ==0) return;
  erow *row = &E.row[E.cy];
  if (E.cx > 0) {
    editorRowDeleteChar(row, E.cx - 1);
    E.cx--;
  } else {
    E.cx = E.row[E.cy - 1].size;
    editorRowAppendString(&E.row[E.cy - 1], row->chars, row->size);
    editorDelRow(E.cy);
    E.cy--;
  }
}

void editorInsertNewline() {
  if (E.cx == 0) {
    editorInsertRow(E.cy, "", 0);
  } else {
    erow *row = &E.row[E.cy];
    editorInsertRow(E.cy + 1, &E.row->chars[E.cx], row->size - E.cx);
    row = &E.row[E.cy];
    row->size = E.cx;
    row->chars[row->size] = '\0';
    editorUpdateRow(row);
  }
  E.cy++;
  E.cx = 0;
}

//
// file io
// -------

char *editorRowsToString(int *buflen) {
  int totlen = 0;
  int j;
  for (j = 0; j < E.numRows; j++)
    totlen += E.row[j].size + 1;
  *buflen = totlen;

  char *buf = malloc(totlen);
  char *p = buf;
  for (j = 0; j < E.numRows; j++) {
    memcpy(p, E.row[j].chars, E.row[j].size);
    p += E.row[j].size;
    *p = '\n';
    p++;
  }

  return buf;
}

void editorOpen(char *filename) {
  free(E.filename);
  E.filename = strdup(filename);

  editorSelectSyntaxHighlight();

  FILE *fp = fopen(filename, "r");
  if (!fp) die("Open file");

  char *line = NULL;
  size_t linecap = 0;
  ssize_t linelen;
  while ((linelen = getline(&line, &linecap, fp)) != -1) {
    while (linelen > 0 && (line[linelen - 1] == '\n' || line[linelen - 1] == '\r'))
      linelen--;
    editorInsertRow(E.numRows, line, linelen);

  }
  free(line);
  fclose(fp);
  E.unsaved = 0;
}

void editorSave(int saveAs) {
  if (E.filename == NULL || saveAs) {
    E.filename = editorPrompt("Save as: %s");
    if (E.filename == NULL) {
      E.filename = NULL;
      editorStatusMessage("Save cancelled");
      return;
    }
    editorSelectSyntaxHighlight();
  }

  int len;
  char *buf = editorRowsToString(&len);

  int fd = open(E.filename, O_RDWR | O_CREAT, 0644);
  if (fd != -1) {
    if (ftruncate(fd, len) != -1) {
      if (write(fd, buf, len) == len) {
        close(fd);
        free(buf);
        E.unsaved = 0;
        editorStatusMessage("%d bytes saved to disk", len);
        return;
      }
    }
    close(fd);
  }
  free(buf);
  editorStatusMessage("Can't save! i/o error: %s", strerror(errno));
}

//
// append buffer
// -------------
struct abuf {
  char *b;
  int len;
};

#define ABUF_INIT {NULL, 0}

void abAppend(struct abuf *ab, const char *s, int len) {
  char *new = realloc(ab->b, ab->len + len);

  if (new == NULL) return;
  memcpy(&new[ab->len], s, len);
  ab->b = new;
  ab->len += len;
}

void abFree(struct abuf *ab) {
  free(ab->b);
}

//
// output
// ------
void editorScroll() {
  E.rx = E.cx;
  if (E.cy < E.numRows) {
    E.rx = editorRowCxToRx(&E.row[E.cy], E.cx);
  }

  if (E.cy < E.rowoff) {
    E.rowoff = E.cy;
  }
  if (E.cy >= E.rowoff + E.screenRows) {
    E.rowoff = E.cy - E.screenRows + 1;
  }
  if (E.rx < E.coloff) {
    E.coloff = E.rx;
  }
  if (E.rx >= E.coloff + E.screenCols) {
    E.coloff = E.rx - E.screenCols + 1;
  }
}

void editorDrawRows(struct abuf *ab) {
  int y;
  for (y = 0; y < E.screenRows; y++) {
    int filerow = y + E.rowoff;
    if (filerow >= E.numRows) {
      if (E.numRows == 0 && y == E.screenRows / 3) {
        char welcome[80];
        int welcomelen = snprintf(welcome, sizeof(welcome), "[ cs edit ] - version %s", CSEDIT_VERSION);
        if (welcomelen > E.screenCols) welcomelen = E.screenCols;
        int padding = (E.screenCols - welcomelen) / 2;
        if (padding) {
          abAppend(ab, "~", 1);
          padding--;
        }
        while (padding--) abAppend(ab, " ", 1); 
        abAppend(ab, welcome, welcomelen);
      } else {
        abAppend(ab, "~", 1);
      }
    } else {
      int len = E.row[filerow].rendersize - E.coloff;
      if (len < 0) len = 0;
      if (len > E.screenCols) len = E.screenCols;
      char *c = &E.row[filerow].render[E.coloff];
      int current_color = -1;
      unsigned char *hl = &E.row[filerow].hl[E.coloff];
      int j;
      for (j = 0; j < len; j++) {
        if (hl[j] == HL_NORMAL) {
          if (current_color != -1) {
            abAppend(ab, "\033[39m", 5);
            current_color = -1;
          }
          abAppend(ab, &c[j], 1);
        } else {
          int color = editorSyntaxToColor(hl[j]);
          if (color != current_color) {
            current_color = color;
            char buf[16];
            int clen = snprintf(buf, sizeof(buf), "\033[%dm", color);
            abAppend(ab, buf, clen);
          }
          abAppend(ab, &c[j], 1);
        }
      }
      abAppend(ab, "\033[39m", 5);
    }

    abAppend(ab, "\033[K", 3);
    abAppend(ab, "\r\n", 2);
  }
}

void editorDrawStatusBar(struct abuf *ab) {
  abAppend(ab, "\033[7m", 4);
  char status[80], rstatus[80];

  char editorMode;
  switch (E.mode) {
    case CommandMode:
      editorMode = 'C';
      break;
    case InputMode:
      editorMode = 'I';
      break;
    case VisualMode:
      editorMode = 'V';
      break;
  }

  int len = snprintf(status, sizeof(status), "%.20s - %d lines %s",
      E.filename ? E.filename : "[No name]", E.numRows, E.unsaved ? "(modified)" : "");
  int rlen = snprintf(rstatus, sizeof(rstatus), "<%s>   %d:%d/%d [%c] ",
      E.syntax ? E.syntax->filetype : "generic" ,E.cx, E.cy + 1, E.numRows, editorMode);
  if (len > E.screenCols) len = E.screenCols;
  abAppend(ab, status, len);
  while (len < E.screenCols) {
    if (E.screenCols - len == rlen) {
      abAppend(ab, rstatus, rlen);
      break;
    } else {
      abAppend(ab, " ", 1);
      len++;
    }
  }
  abAppend(ab, "\033[m", 4);
  abAppend(ab, "\r\n", 2);
}

void editorDrawMsgBar(struct abuf *ab) {
  abAppend(ab, "\033[K", 3);
  int msglen = strlen(E.statusmsg);
  if (msglen > E.screenCols) msglen = E.screenCols;
  if (msglen && time(NULL) - E.statusmsg_time < 5)
    abAppend(ab, E.statusmsg, msglen);
}

void editorRefreshScreen() {
  editorScroll();

  struct abuf ab = ABUF_INIT;

  abAppend(&ab, "\033[?25l", 6);
  abAppend(&ab, "\033[H", 3);
  editorDrawRows(&ab);
  editorDrawStatusBar(&ab);
  editorDrawMsgBar(&ab);

  char buf[32];
  snprintf(buf, sizeof(buf), "\033[%d;%dH", (E.cy - E.rowoff) + 1, (E.rx - E.coloff) + 1);
  abAppend(&ab, buf, strlen(buf));

  abAppend(&ab, "\033[?25h", 6);

  write(STDOUT_FILENO, ab.b, ab.len);
  abFree(&ab);
}

void editorStatusMessage(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  vsnprintf(E.statusmsg, sizeof(E.statusmsg), fmt, ap);
  va_end(ap);
  E.statusmsg_time = time(NULL);
}

void editorPrevWord() {
  int x = E.cx;
  int y = E.cy;
  int moved = 0;
  erow *row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];

  while (row) {
    if (moved && row->chars[x] != ' ' && row->chars[x] != '\t' && (x == 0 || row->chars[x - 1] == ' ' || row->chars[x - 1] == '\t')) {
      // EXIT cond
      E.cx = x;
      E.cy = y;
      break;
    } else if (x == 0) {
      y--;
      row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];
      if (!row) return;
      x = row->size;
      moved++;
    } else {
      x--;
      moved++;
    }
  }
}

void editorNextWord() {
  int ws = 0;
  int x = E.cx;
  int y = E.cy;
  erow *row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];

  while (row) {
    if (ws > 0 && row->chars[x] != ' ' && row->chars[x] != '\t') {
      E.cx = x;
      E.cy = y;
      break;
    } else if (row->chars[x] == ' ' || row->chars[x] == '\t') {
      x++;
      ws++;
    } else if (x + 1 >= row->size) {
      y++;
      x = 0;
      ws++;
      row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];
      if (!row) return;
    } else {
      x++;
    }
  }
}

void editorEndOfWord() {
  int x = E.cx;
  int y = E.cy;
  erow *row = (y >= E.numRows) ? NULL : &E.row[y];

  while (1) {
    if (row->chars[x] != ' ' && row->chars[x] != '\t' && (x + 1 >= row->size || (row->chars[x + 1] == ' ' || row->chars[x + 1] == '\t'))) {
      E.cx = x;
      E.cy = y;
      break;
    } else {
      x++;
    }
  }
}

//
// input
// -----
char *editorPrompt(char *prompt) {
  size_t bufsize = 128;
  char *buf = malloc(bufsize);

  size_t buflen = 0;
  buf[0] = '\0';

  while (1) {
    editorStatusMessage(prompt, buf);
    editorRefreshScreen();

    int c = editorReadKey();
    if (c == DEL_KEY || c == CTRL_KEY('h') || c == BACKSPACE) {
      if (buflen != 0) buf[--buflen] = '\0';
    } else if (c == '\033') {
      editorStatusMessage("");
      free(buf);
      return NULL;
    } else if (c == '\r') {
      if (buflen != 0) {
        editorStatusMessage("");
        return buf;
      } else {
        editorStatusMessage("");
        free(buf);
        return NULL;
      }
    } else if (!iscntrl(c) && c < 128) {
      if (buflen == bufsize - 1) {
        bufsize *= 2;
        buf = realloc(buf, bufsize);
      }
      buf[buflen++] = c;
      buf[buflen] = '\0';
    }
  }
}

void editorMoveCursor(int key) {
  erow *row = (E.cy >= E.numRows) ? NULL : &E.row[E.cy];
  switch (key) {
    case 'j':
    case ARROW_DOWN:
      if (E.cy < E.numRows) {
        E.cy++;
      }
      break;
    case 'k':
    case ARROW_UP:
      if (E.cy != 0) {
        E.cy--;
      }
      break;
    case 'h':
    case ARROW_LEFT:
      if (E.cx != 0) {
        E.cx--;
      }
      break;
    case 'l':
    case ARROW_RIGHT:
      {
        int mod = E.mode == CommandMode ? 1 : 0;
        if (row && E.cx < row->size - mod)
          E.cx++;
        break;
      }
  }

  row = (E.cy >= E.numRows) ? NULL : &E.row[E.cy];
  int rowlen = row ? row->size : 0;
  if (E.cx > rowlen)
    E.cx = rowlen;
}

void editorPageCursor(int key) {
  if (key == PAGE_DOWN) {
    if (E.numRows < E.screenRows) {
      E.cy = E.numRows - 1;
      return;
    }
    E.rowoff = E.rowoff + E.screenRows;
    if (E.rowoff > E.numRows - E.screenRows)
      E.rowoff = E.numRows - E.screenRows;
    E.cy = E.rowoff + 1;
  } else if (key == PAGE_UP) {
    if (E.cy < E.screenRows) {
      E.cy = 0;
      return;
    }
    E.rowoff = E.rowoff - E.screenRows;
    if (E.rowoff < 0)
      E.rowoff = 0;
    E.cy = E.rowoff + E.screenRows - 1;
  }

  erow *row = (E.cy >= E.numRows) ? NULL : &E.row[E.cy];
  int rowlen = row ? row->size : 0;
  if (E.cx > rowlen)
    E.cx = rowlen;
}

void editorInputKp(int c) {
  switch (c) {
    case '\033':
    case CTRL_KEY('l'):
      if (E.cx == E.row[E.cy].size) editorMoveCursor(ARROW_LEFT);
      E.mode = CommandMode;
      break;
    case '\r':
      editorInsertNewline();
      break;
    case BACKSPACE:
    case DEL_KEY:
    case CTRL_KEY('h'):
      if (c == DEL_KEY) editorMoveCursor(ARROW_RIGHT);
      editorDeleteChar();
      break;
    default:
      editorInsertChar(c);
      break;
  }
}

void editorCommandKp(int c) {
  static int deleting = 0;
  static int counter = 1;
  static int place = 1;
  switch (c) {
    case 'j':
    case 'k':
    case 'h':
    case 'l':
      while (counter > 0) {
        editorMoveCursor(c);
        counter--;
      }
      break;
    case 'g':
      E.cy = 0;
      break;
    case 'G':
      if (E.numRows > E.screenRows) {
        E.cy = E.numRows - 1;
        E.rowoff = E.numRows - E.screenRows - 1;
      } else {
        E.cy = E.numRows - 1;
      }
      break;
    case 'o':
      editorCommandKp('$');
      editorCommandKp('i');
      editorMoveCursor(ARROW_RIGHT);
      editorInsertNewline();
      break;
    case 'O':
      editorCommandKp('^');
      editorInsertNewline();
      editorMoveCursor(ARROW_UP);
      editorCommandKp('i');
      break;
    case 'x':
      while (counter > 0) {
        E.mode = InputMode;
        editorMoveCursor(ARROW_RIGHT);
        E.mode = CommandMode;
        editorDeleteChar();
        if (E.cx == E.row[E.cy].size) editorMoveCursor(ARROW_LEFT);
      }
      break;
    case 'i':
    case 'a':
      E.mode = InputMode;
      if (c == 'a') editorMoveCursor('l');
      break;
    case 'd':
      if (!deleting) {
        deleting++;
        return;
      } else {
        while (counter > 0) {
          editorDelRow(E.cy);
          if (E.cy == E.numRows) E.cy--;
          if (E.cy < 0) E.cy = 0;
          counter--;
        }
        break;
      }
    case '$':
      if (deleting) {
        int diff = E.row[E.cy].size - E.cx;
        E.cx = E.row[E.cy].size - 1;
        if (E.cx < 0) E.cx = 0;
        while (diff > 0) {
          E.mode = InputMode;
          editorMoveCursor(ARROW_RIGHT);
          E.mode = CommandMode;
          editorDeleteChar();
          if (E.cx == E.row[E.cy].size) editorMoveCursor(ARROW_LEFT);
          diff--;
        }
      } else if (E.cy < E.numRows) {
        E.cx = E.row[E.cy].size - 1;
        if (E.cx < 0) E.cx = 0;
      }
      break;
    case 'w':
      while (counter > 0) {
        counter--;
        if (deleting) {
          int orig = E.cx;
          editorNextWord();
          int diff = E.cx - orig;
          if (diff < 0) {
            editorPrevWord();
            editorEndOfWord();
            editorMoveCursor(ARROW_RIGHT);
            diff = E.cx - orig;
          }
          while (diff > 0) {
            editorDeleteChar();
            diff--;
          }
        } else {
          editorNextWord();
        }
      }
      break;
    case 'b':
      while (counter > 0) {
        if (deleting) {
          // TODO
        } else {
          editorPrevWord();
        }
        counter--;
      }
      break;
    case 'e':
      while (counter > 0) {
        if (deleting) {
          // TODO
        } else {
          editorEndOfWord();
        }
        counter--;
      }
      break;
    case '^':
      if (deleting) {
        while (E.cx > 0) {
          editorDeleteChar();
        }
      } else {
        E.cx = 0;
      }
      break;
    case '0':
      if (counter == 1) {
        if (deleting) {
          while (E.cx > 0) {
            editorDeleteChar();
          }
        } else {
          E.cx = 0;
        }
        break;
      }
      /* fall through */
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      {
        if (counter <= 1 && place == 1) counter--;
        int num = c - 48;
        counter += num * place;
        place *= 10;
      }
      return;
  }
  deleting = 0;
  place = 1;
  counter = 1;
}

void editorProcessKeypress() {
  static int quit_times = UNSAVED_QUIT_COUNT;
  int c = editorReadKey();
  switch (c) {
    case CTRL_KEY('q'):
      if (E.unsaved && quit_times > 0) {
        editorStatusMessage("Warning: unsaved changed. Press Ctrl-q %d more times to quit anyway", quit_times);
        quit_times--;
        return;
      }
      quit();
      break;
    case ARROW_DOWN:
    case ARROW_UP:
    case ARROW_LEFT:
    case ARROW_RIGHT:
      editorMoveCursor(c);
      return;
    case PAGE_UP:
    case PAGE_DOWN:
      editorPageCursor(c);
      return;
    case CTRL_KEY('u'):
      editorPageCursor(PAGE_UP);
      return;
    case CTRL_KEY('d'):
      editorPageCursor(PAGE_DOWN);
      return;
    case CTRL_KEY('s'):
      editorSave(0);
      return;
    case CTRL_KEY('a'):
      editorSave(1);
      return;
    case HOME_KEY:
      E.cx = 0;
      break;
    case END_KEY:
      if (E.cy < E.numRows)
        E.cx = E.row[E.cy].size;
      break;
  }

  quit_times = UNSAVED_QUIT_COUNT;

  if (E.mode == CommandMode) {
    editorCommandKp(c);
  } else {
    editorInputKp(c);
  }
}

//
// init
// ----
void initEditor() {
  E.mode           = CommandMode;
  E.rowoff         = 0;
  E.coloff         = 0;
  E.numRows        = 0;
  E.unsaved        = 0;
  E.statusmsg_time = 0;
  E.cx             = 0;
  E.cy             = 0;
  E.rx             = 0;
  E.row            = NULL;
  E.filename       = NULL;
  E.syntax         = NULL;
  E.statusmsg[0]  = '\0';

  if (getWindowSize(&E.screenRows, &E.screenCols) == -1) die("getWindowSize");
  E.screenRows -= 2;
}

int main(int argc, char *argv[]) {
  enableRawMode();
  initEditor();
  if (argc >= 2) {
    editorOpen(argv[1]);
  }

  editorStatusMessage("Ctrl-s to save - Ctrl-q to quit");

  while (1) {
    editorRefreshScreen();
    editorProcessKeypress();
  }
  return 0;
}
